@@ -1,6 +1,9 @@
+import pickle
 import sys
 import os
 import pathlib
+import shutil
+
 
 from mlflow.tracking import MlflowClient
 import mlflow.pyfunc
@@ -32,19 +35,22 @@ import traceback
 
 
 from .config import p as cfg
-from .basehandler import BaseHandler, load
-from .basehandler import load as load_BaseHandler
+from .basehandler import BaseHandler
+#from .basehandler import load as load_BaseHandler
 import atexit
+import glob
 
-__version__ = "0.1.20"
+__version__ = "0.2.01"
 _eureka_client = None
 
+
 @atexit.register
 def cleanup():
     global _eureka_client
     if _eureka_client:
         _eureka_client.stop()
 
+
 class Server:
 
     def __init__(self, config=None):
@@ -84,7 +90,7 @@ class Server:
                 if _eureka_client:
                     print("Eureka client already started")
                 else:
-                    
+
                     _eureka_client = eureka_client.EurekaClient(
                         eureka_server=self.config.eureka_server,
                         app_name=self.config.app_name,
@@ -100,10 +106,6 @@ class Server:
                     "message": str(ex),
                 }
 
-        # init schedulerr
-        self.scheduler = None
-        self.init_scheduler()
-
         # init FastAPI app
         tags_metadata = [
             {
@@ -134,7 +136,7 @@ class Server:
             except:
                 pid = os.getpid()
                 os.kill(pid, 9)
-            
+
             return self.app.openapi_schema
         self.app.openapi = custom_openapi
 
@@ -157,7 +159,7 @@ class Server:
         self.app.routes.append(starlette_Route(
             self.config.basepath+"/", redirect_to_docs))
         self.app.routes.append(starlette_Route(
-            self.config.basepath+"/info", redirect_to_docs))                        
+            self.config.basepath+"/info", redirect_to_docs))
 
         # create model list endpoint
         @self.app.get(
@@ -187,7 +189,7 @@ class Server:
             tags=["Metadata"],
             description="Get a dict of all errors.",
             response_model=Dict[str, Any],
-            include_in_schema=False
+            include_in_schema=True
         )
         async def errors():
             return self.error_dict
@@ -203,7 +205,7 @@ class Server:
             self.init_scheduler()
             return "OK. Please be patient :)"
 
-        #dummy route for the  eureka health call
+        # dummy route for the  eureka health call
         if self.config.eureka_server:
             @self.app.get(
                 self.config.basepath+'/health',
@@ -213,25 +215,27 @@ class Server:
             )
             async def health():
                 return {}
+
             @self.app.post(
                 self.config.basepath+'/health',
                 tags=["Metadata"],
                 description="Eureka health call.",
                 include_in_schema=False
             )
-            async def health():
+            async def health2():
                 return {}
 
-        # create model artifact endpoints
-
         # check caching
-        if self.config.cache:
-            self.full_cache_dir = os.path.join(
-                pathlib.Path().absolute(), self.config.cachedir)
-            logger.info(f"Use cachedir: {self.full_cache_dir}")
-            pathlib.Path(self.full_cache_dir).mkdir(
-                parents=True, exist_ok=True)
-            self.update_models_from_cache()
+        self.full_cache_dir = os.path.join(
+            pathlib.Path().absolute(), self.config.cachedir)
+        logger.info(f"Use cachedir: {self.full_cache_dir}")
+        pathlib.Path(self.full_cache_dir).mkdir(
+            parents=True, exist_ok=True)
+        self.update_models_from_cache()
+
+        # init schedulerr
+        self.scheduler = None
+        self.init_scheduler()
 
     def check_token(self, token):
         """
@@ -262,30 +266,37 @@ class Server:
     def update_models_from_cache(self):
         logger.info(f"Update models from cache")
         # load all available cached models
-        for filename in os.listdir(self.full_cache_dir):
-            newmodel = load_BaseHandler(os.path.join(
-                self.full_cache_dir, filename), self)
-            self.model_dict[newmodel.name] = newmodel
+        meta_file = os.path.join(self.full_cache_dir, "meta.pkl")
+        if os.path.exists(meta_file):
+            with open(meta_file, "rb") as f:
+                old_config = pickle.load(f)
+                for name, m in old_config.items():
+                    try:
+                        newmodel = BaseHandler(
+                            self, m["model"], m["model_version"])
+                        newmodel.register_route()
+                        self.model_dict[name] = newmodel
+                    except Exception as ex:
+                        self.error_dict[name] = {
+                            "message": str(ex),
+                            "type": "Exception"
+                        }
 
         self.app.openapi_schema = None
 
     def update_models(self):
         logger.info(f"Update models")
 
+        # get a list of the latest models
         all_models = []
         try:
             all_models = self.client.list_registered_models()
             if "server" in self.error_dict:
                 del self.error_dict["server"]
         except Exception as ex:
-            exc_type, exc_obj, exc_tb = sys.exc_info()
-            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
             self.error_dict["server"] = {
                 "message": str(ex),
-                "type": str(exc_type),
-                "file": fname,
-                "line": exc_tb.tb_lineno,
-                "exception": exc_obj
+                "type": "Exception"
             }
 
         for m in all_models:
@@ -298,7 +309,9 @@ class Server:
 
             # if the currently loaded model is already ok
             if name in self.model_dict and model_version.run_id == self.model_dict[name].run_id:
-                continue
+                # check if the model is still working
+                if self.model_dict[name].health():
+                    continue
 
             # check if the type tag fits
             if len(self.config.tags) > 0 and all([not tt in m.tags.keys() for tt in self.config.tags]):
@@ -306,42 +319,64 @@ class Server:
 
             logger.info(f"Update model {name}")
 
-            # delete old route
-            for idx2 in [idx for idx, r in enumerate(self.app.routes) if r.path == self.config.basepath+"/"+name]:
-                del self.app.routes[idx2]
-
             # create new handler
             try:
                 newmodel = BaseHandler(self, m, model_version)
+
+                # delete old route
+                for idx2 in [idx for idx, r in enumerate(self.app.routes) if r.path == self.config.basepath+"/"+name]:
+                    del self.app.routes[idx2]
+
+                # register new route
+                newmodel.register_route()
+
+                # keep old model
+                if name in self.model_dict:
+                    oldmodel = self.model_dict[name]
+                else:
+                    oldmodel = None
+
                 self.model_dict[name] = newmodel
 
                 if name in self.error_dict:
                     del self.error_dict[name]
 
-                if self.config.cache:
-                    # create a cache representation of the new model
-                    newmodel.save(os.path.join(self.full_cache_dir, name))
+                with open(os.path.join(self.full_cache_dir, "meta.pkl"), "wb") as f:
+                    pickle.dump({
+                        key: {
+                            "model": val.m,
+                            "model_version": val.model_version
+                        }
+                        for key, val in self.model_dict.items()
+                    }, f)
+
+                # finally stop the old mlflow server
+                if oldmodel is not None:
+                    oldmodel._stop_server()
+                    del oldmodel
 
             except Exception as ex:
-                exc_type, exc_obj, exc_tb = sys.exc_info()
-                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                 self.error_dict[name] = {
                     "message": str(ex),
-                    "type": str(exc_type),
-                    "file": fname,
-                    "line": exc_tb.tb_lineno,
-                    "exception": exc_obj,
-                    "traceback": traceback.format_exc().split("\n"),
+                    "type": "Exception"
                 }
 
-        self.app.openapi_schema = None
+        # cleanup non registerd models
 
-    def load_artifact(self, run_id, artifact_path):
-        """ load an artifact from server, used for input_example
-        """
-        params = {"path": artifact_path, "run_uuid": run_id}
-        res = requests.get(self.config.mlflow+"/get-artifact", params=params)
-        return res
+        # cleanup cache
+        for folder in glob.glob(os.path.join(pathlib.Path().absolute(), self.config.cachedir, "*")):
+            if os.path.isdir(folder):
+                run_id = os.path.basename(folder)
+                usefull = any(
+                    [any([v.run_id == run_id for v in m.latest_versions]) for m in all_models])
+                if usefull == False:
+                    dir_path = pathlib.Path(folder)
+                    try:
+                        shutil.rmtree(dir_path)
+                    except Exception:
+                        pass
+
+        self.app.openapi_schema = None
 
     def init_scheduler(self):
         """ init and restart the scheduler
